﻿using System;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading;
using RedRat.AvDeviceDb;
using RedRat.RedRat3;
using RedRat.RedRat4.StreamIO.X;
using RedRat.Usb;
using RedRat.Util;
using RedRat.Util.Serialization;

namespace RedRat.Samples.KeyPressLogger
{
    /// <summary>
    /// This application will capture incoming IR signals from a USB RedRat device (RedRat3-II or RedRat-X),
    /// attempt to decode them and the print out the results.
    /// 
    /// Windows: To send the log output to a file, use this:
    ///     KeyPressLogger.exe "RedRat-X 1" DeviceDB.xml > test.log
    /// 
    /// Windows: Using Powershell, output can be sent to the console AND log file like this:
    ///     KeyPressLogger.exe "RedRat-X 1" DeviceDB.xml | Tee-Object -file test.log
    ///
    /// Linux: To run and log output:
    ///     dotnet KeyPressLogger.dll RedRatX-1 ./DeviceDB.xml > keys.log
    /// </summary>
    public class Program
    {
        private IRedRat3 redRat;
        private AVDeviceDB avDeviceDB;
        private string dbFilename;
        private bool oneshot = false;
        private readonly ManualResetEvent completionWaitHandle = new ManualResetEvent( false );

        static void Main( string[] args )
        {
            new Program( args );
        }

        public Program( string[] args )
        {
            try
            {
                CheckArgs( args );
            }
            catch ( Exception ex )
            {
                Console.Error.WriteLine( $"\nError from Key Press Logger: {ex.Message}\n" );
            }
            finally
            {
                TidyUp();
            }
        }

        private void CheckArgs( string[] args )
        {
            // Need two - RR3 to use and device DB file.
            if ( args.Length < 2 )
            {
                PrintUsage();
                return;
            }

            if ( !FindUsbRedRats( args[0] ) ) return;
            if ( !LoadSignalData( args[1] ) ) return;

            if ( args.Length > 2 )
            {
                if ( args[2].Equals( "-oneshot", StringComparison.CurrentCultureIgnoreCase ) )
                {
                    oneshot = true;
                }
            }

            avDeviceDB.DecoderLogEnabled = false; // false = default. Set to true to turn on some decoder logging.
            StartLogging();
        }

        /// <summary>
        /// Handles the user pressing CTRL-C to exit the application. If the RC detector is enabled
        /// when the application stops it can cause the RR3 to hang.
        /// </summary>
        private void ConsoleCancelKeyPress( object sender, ConsoleCancelEventArgs e )
        {
            TidyUp();
        }

        private bool FindUsbRedRats( string rrName )
        {
            // Lookup using the 'friendly' name, e.g. "RedRatX-0".
            var usbRrs = RRUtil.FindRedRats( LocationInfo.RedRatType.USB ).Cast<UsbLocationInfo>();

            var li = usbRrs.FirstOrDefault( r => r.Name.Equals( rrName, StringComparison.CurrentCultureIgnoreCase ) );
            if ( li != null )
            {
                redRat = LocationInfo.IsRedRatX(li.RRType) ? RedRatX.GetInstance(li) : li.GetRedRat() as IRedRat3;
                return true;
            }

            // OK, haven't found an RRX so try using the device name.
            li = usbRrs.FirstOrDefault( r => r.DeviceName.Equals( rrName, StringComparison.CurrentCultureIgnoreCase ) );
            if ( li != null )
            {
                redRat = LocationInfo.IsRedRatX( li.RRType ) ? RedRatX.GetInstance( li ) : li.GetRedRat() as IRedRat3;
                return true;
            }

            WriteError($"Failed to find to RedRat-X '{rrName}'");
            PrintUsage();
            return false;
        }

        private bool LoadSignalData( string fileName )
        {
            try
            {
                avDeviceDB = Serializer.AvDeviceDbFromXmlFile(fileName).Object;
                dbFilename = fileName;
            }
            catch ( Exception ex )
            {
                WriteError( $"Error loading signal database file '{fileName}'\n   {ex.Message}" );
                return false;
            }
            return true;
        }

        private void LogEvent( string eventMsg )
        {
            // Prints out to 100ths of a second.
            Console.WriteLine( "[{0:dd/MM/yyyy HH:mm:ss.ff}] {1}", DateTime.Now, eventMsg );
        }

        private void PrintUsage()
        {
            var version = Assembly.GetExecutingAssembly().GetName().Version;
            Console.WriteLine( "\nRedRat Key Press Logger - Version {0}.{1}", version.Major, version.Minor );
            Console.WriteLine( "-------------------------------------" );
            Console.WriteLine( "\nUsage: KeyPressLogger RedRatX-Name IR-Signal-DB-File [-oneshot]" );
            Console.WriteLine( "\n  Where:\n" );
            Console.WriteLine( "    RedRatX-Name      - Either the friendly name or USB device name." );
            Console.WriteLine( "    IR-Signal-DB-File - File containing IR signal XML data for decoding." );
            Console.WriteLine( "    -oneshot          - [Optional] Program will terminate after IR signal input.\n");

            Console.WriteLine( PrintUsbRedRats() );
        }

        /// <summary>
        /// Handles incoming IR signal events from RR3.
        /// </summary>
        private void RedRatXOnRcDetectorSignalIn( object sender, EventArgs eventArgs )
        {
            try
            {
                var siea = eventArgs as SignalEventArgs;
                if ( siea == null ) return;

                switch ( siea.Action )
                {
                    case SignalEventAction.RC_DETECTOR_ENABLED:
                        LogEvent( "Signal input enabled" );
                        break;

                    case SignalEventAction.RC_DETECTOR_DISABLED:
                        LogEvent( "Signal input terminated" );
                        break;

                    case SignalEventAction.MODULATED_SIGNAL:
                        var sigKey = avDeviceDB.DecodeSignal( siea.ModulatedSignal );
                        var msg = sigKey.KeyStatus == SignalKey.Status.KNOWN
                            ? $"KEY: {sigKey.AVDevice.Name} -> {sigKey.Signal.Name}"
                            : "KEY: Unknown";
                        LogEvent( msg );
                        completionWaitHandle.Set(); // For use if "oneshot"
                        break;

                    case SignalEventAction.EXCEPTION:
                        LogEvent( "Error: " + siea.Exception.Message );
                        break;
                }
            }
            catch ( Exception ex )
            {
                Console.Error.WriteLine( "Error decoding IR signal:\n" + ex );
            }
        }

        private void StartLogging()
        {
            // Setup CTRL-C handler.
            Console.CancelKeyPress += ConsoleCancelKeyPress;

            // Connect to the RedRat (RedRat3 doesn't need this, but no harm.)
            redRat.Connect();

            // Enable IR input...
            redRat.RCDetectorSignalIn += RedRatXOnRcDetectorSignalIn;

            redRat.RCDetectorEnabled = true;

            if ( oneshot )
            {
                Console.WriteLine("\nRedRat Key Logger started. Waiting for one IR signal only.");

                // Wait for one signal only
                completionWaitHandle.WaitOne();
            }
            else
            {
                Console.WriteLine( "\nRedRat Key Logger Started...\n" );
                Console.WriteLine( $"   RedRat            : {redRat.LocationInformation.Name}" );
                Console.WriteLine( $"   Signal DB Filename: {dbFilename}\n" );
                Console.WriteLine( " Hit <RETURN> to terminate...\n" );
                Console.ReadLine();
            }
        }

        private void TidyUp()
        {
            if ( redRat == null ) return;
            if ( redRat.RCDetectorEnabled )
            {
                redRat.RCDetectorEnabled = false;
            }
            redRat.Disconnect();
            RRUtil.DisposeOfAllRedRats();
        }

        private void WriteError( string message )
        {
            var currentColor = Console.ForegroundColor;
            Console.ForegroundColor = ConsoleColor.Red;
            Console.Error.WriteLine( message );
            Console.ForegroundColor = currentColor;
        }

        private string PrintUsbRedRats()
        {
            var sb = new StringBuilder();
            var usbRrs = RRUtil.FindRedRats(LocationInfo.RedRatType.USB).Cast<UsbLocationInfo>();
            if (usbRrs.Any())
            {
                sb.Append("  Have the following USB RedRat devices:\n");
                foreach (var li in usbRrs)
                {
                    sb.Append($"    Device={li.DeviceName}, Name={li.Name}, Type={li.RRType}\n");
                }
            }
            else
            {
                sb.Append("  No USB RedRat devices found.\n");
            }

            return sb.ToString();
        }
    }
}
