using System;
using System.Linq;
using System.Threading;
using System.Reflection;
using System.Xml.Serialization;
using RedRat.IR;
using RedRat.Util;
using RedRat.RedRat3;
using RedRat.RedRat3.USB;
using RedRat.RedRat4.StreamIO.X;
using RedRat.Usb;
using RedRat.Util.Serialization;

namespace RedRat.Samples.RR_CLI
{

    /// <summary>
    /// Command line interface for basic operations using USB RedRat devices. Used for scripting etc.
    /// </summary>
    class RR_CLI : IXmlSerializerHandler
    {

        /// <summary>
        /// The RedRat we use for operations.
        /// </summary>
        private IRedRat3 redRat = null;

        private readonly ArgumentParser argParser = new ArgumentParser();

        private bool haveSignal = false; // Set true once signal (or ex) input

        private IRPacket irPacket = null;

        /// <summary>
        /// The main entry point for the application.
        /// </summary>
        [STAThread]
        static void Main( string[] args )
        {
            // If calling from a web server etc. - need to ensure that it does not need access to the desktop.
            RedRat3USBImpl.IS_SERVER_PROCESS = true;
            new RR_CLI( args );
        }

        public RR_CLI( string[] args )
        {
            try
            {
                if ( argParser.ParseArgs( args, out var message ) )
                {
                    if ( GetRedRat() )
                    {
                        Run();
                    }
                    else
                    {
                        Console.WriteLine( $"Failed to connect to RedRat {argParser.RedRatName}" );
                        Console.WriteLine( RRUtil.PrintFoundRedRats() );
                    }
                }
                else
                {
                    PrintUsage( message );
                }
            }
            catch ( Exception ex )
            {
                Console.WriteLine( "\nERROR: " + ex.Message );
                Console.WriteLine( ex.StackTrace + "\n" );
            }
        }

        private bool GetRedRat()
        {
            // Find locationInfo for each USB RedRat (RedRat3 and RedRat-X)
            var locInfos = RRUtil.FindRedRats( LocationInfo.RedRatType.USB ).OfType<UsbLocationInfo>();

            // Get the device with the given name or device name
            var rrLi = locInfos.FirstOrDefault( li => li.Name == argParser.RedRatName || li.DeviceName == argParser.RedRatName );

            // Slightly cryptic, but if:
            //  - RedRat3: The LocationInfo will give the RedRat device object directly.
            //  - RedRatX: The LocationInfo returns an intermediate object (due to the way the full RedRatX object abstracts network
            //             or USB comms), so have to use static method on RedRatX.
            redRat = ( rrLi is RedRatXUsbLocationInfo ? RedRatX.GetInstance( rrLi ) : rrLi?.GetRedRat() ) as IRedRat3;

            return redRat != null;
        }

        private void Run()
        {
            // An RR3 doesn't actually need a connect call, but makes the code work with both RR3 & RRX.
            redRat.Connect();
            try
            {
                if ( argParser.ShouldBlink )
                {
                    redRat.Blink();
                    Console.WriteLine( "BLINK" );
                }

                if ( argParser.ShouldCapture )
                {
                    CaptureSignal();
                }

                if ( argParser.ShouldOutput )
                {
                    OutputSignal();
                }
            }
            finally
            {
                redRat.Disconnect();
            }
        }

        private bool IsXmlFile( string fileName )
        {
            var extension = fileName.Substring( startIndex: fileName.Length - 3 );
            return string.Equals( extension, "xml", StringComparison.InvariantCultureIgnoreCase );
        }

        private void CaptureSignal()
        {
            redRat.LearningSignalIn += SignalDataHandler;
            redRat.GetModulatedSignal( 10000 ); // 10s timeout

            // Wait until have signal...
            Console.WriteLine( "Waiting (10s) for signal input...." );
            haveSignal = false;
            while ( !haveSignal )
            {
                Thread.Sleep( 10 );
            }

            if ( irPacket != null )
            {
                var fileName = argParser.GetArgOption( ArgumentParser.CaptureArg );
                if ( IsXmlFile( fileName ) )
                {
                    Serializer.IrPacketToXmlFile( irPacket, fileName );
                    Console.WriteLine( "Saved in XML format to file: " + fileName );
                }
                else
                {
                    Serializer.IrPacketToBinaryFile( irPacket, fileName );
                    Console.WriteLine( "Saved in binary format to file: " + fileName );
                }
            }
        }

        private IRPacket LoadXmlSignal( string fileName )
        {
            var device = argParser.HaveArg( ArgumentParser.DeviceArg ) ? argParser.GetArgOption( ArgumentParser.DeviceArg ) : null;
            var signal = argParser.HaveArg( ArgumentParser.SignalArg ) ? argParser.GetArgOption( ArgumentParser.SignalArg ) : null;

            var result = XmlSignalLoader.LoadSignal( fileName, device, signal, out var message );

            if ( result == null )
            {
                Console.WriteLine( message );
            }

            return result;
        }

        private void OutputSignal()
        {
            var fileName = argParser.GetArgOption( ArgumentParser.OutputArg );
            if ( IsXmlFile( fileName ) )
            {
                irPacket = LoadXmlSignal( fileName );
                if ( irPacket == null ) return;
            }
            else
            {
                irPacket = Serializer.IrPacketFromBinaryFile( fileName );
            }

            // Print signal if requested
            if ( argParser.HavePrintArg )
            {
                Console.WriteLine( irPacket );
            }

            // Output signal
            if ( irPacket is ModulatedSignal )
            {
                redRat.OutputModulatedSignal( irPacket );
                Console.WriteLine( "Output modulated signal." );
            }
            else if ( irPacket is DoubleSignal )
            {
                // Only take first signal
                var sig1 = ( (DoubleSignal) irPacket ).Signal1;
                if ( sig1 != null )
                {
                    redRat.OutputModulatedSignal( sig1 );
                    Console.WriteLine( "Output modulated signal (sig1)." );
                }
            }
        }

        private void PrintUsage( string errorMessage )
        {
            // Get version info.
            var assembly = Assembly.GetExecutingAssembly();

            var ata = (AssemblyTitleAttribute[]) assembly.GetCustomAttributes( typeof( AssemblyTitleAttribute ), true );
            var version = assembly.GetName().Version;
            var headerLine = $"{ata[0].Title} V{version} - Command line interface to the RedRat.";

            Console.WriteLine( headerLine );
            Console.WriteLine( "-------------------------------------------------------\n" );
            Console.WriteLine( errorMessage );
            Console.WriteLine( "" );
            Console.WriteLine( "  Usage: RR_CLI <RedRat name> <options>\n" );
            Console.WriteLine( "  Options are:" );
            Console.WriteLine( "     -blink              Blinks the red LED." );
            Console.WriteLine( "     -capture filename   Captures signal and stores in file <filename>." );
            Console.WriteLine( "     -output filename    Outputs a signal stored in file <filename>." );
            Console.WriteLine( "     -print              Prints the IR signal data when output." );
            Console.WriteLine( "     -device devicename  Specifies the device to use when working with an xml" );
            Console.WriteLine( "                         signal database. (Use -output to set the file.)" );
            Console.WriteLine( "     -signal signalname  Specifies the signal to use when working with an xml" );
            Console.WriteLine( "                         signal database or xml device file. (Use -output to" );
            Console.WriteLine( "                         set the file.)" );
            Console.WriteLine( "" );
            Console.WriteLine( RRUtil.PrintFoundRedRats( LocationInfo.RedRatType.USB ) );
            Console.WriteLine( "Press RETURN to continue..." );
            Console.ReadLine();
        }

        /// <summary>
        /// Handles input signal (or error) from the RedRat.
        /// </summary>
        private void SignalDataHandler( object sender, EventArgs e )
        {

            if ( e is SignalEventArgs siea )
            {
                switch ( siea.Action )
                {
                    case SignalEventAction.EXCEPTION:
                        Console.WriteLine( siea.Exception.Message );
                        break;

                    case SignalEventAction.MODULATED_SIGNAL:
                        Console.WriteLine( "Have IR data..." );
                        irPacket = siea.ModulatedSignal;
                        break;

                    default:
                        Console.WriteLine( "Unknown return..." );
                        break;
                }

                haveSignal = true;
            }
            else
            {
                Console.WriteLine( "Event of unknown type...." );
            }
        }

        #region IXmlSerializerHandler Interface

        public void UnknownNode( object sender, XmlNodeEventArgs e )
        {
            Console.WriteLine( "Error parsing file - unknown Node: " + e.Name + "\t" + e.Text );
        }

        public void UnknownAttribute( object sender, XmlAttributeEventArgs e )
        {
            System.Xml.XmlAttribute attr = e.Attr;
            Console.WriteLine( "Error parsing file - unknown attribute " + attr.Name + "='" + attr.Value + "'" );
        }

        #endregion
    }
}
