﻿using RedRat.IR;
using RedRat.RedRat3;
using RedRat.Util;
using System;
using RedRat.AvDeviceDb;
using RedRat.Util.Serialization;

namespace RedRat.Samples.SimpleSignalOutput
{
    /// <summary>
    /// This sample shows how to read in a signal database XML file, extract one signal from it
    /// and then output via a USB RedRat device.
    /// 
    /// The database file, device and signal names are given as command line parameters:
    /// 
    ///     SimpleSignalOutput database-file.xml device-name signal-name
    /// e.g. 
    ///     SignalSignalOutput DeviceDB.xml CD Play
    /// 
    /// It is not an efficient way of sending sequences of IR signals as loading the XML file
    /// for each signal is quite an expensive operation.
    /// </summary>
    public class SimpleSignalOutput
    {
        static void Main(string[] args)
        {
            new SimpleSignalOutput( args );
        }

        public SimpleSignalOutput( string[] args )
        {
            try
            {
                CheckArgs( args );

                using var rr3 = FindRedRat();

                // An RR3 doesn't actually need a connect call, but makes the code work with both RR3 & RRX.
                rr3.Connect();

                var signalDB = Serializer.AvDeviceDbFromXmlFile( args[0] ).Object;
                var signal = GetSignal( signalDB, args[1], args[2] );
                rr3.OutputModulatedSignal( signal );
                Console.WriteLine( $"\nSignal {args[1]}->{args[2]} output via {rr3.LocationInformation.Name}." );

                rr3.Disconnect();
            }
            catch ( Exception ex )
            {
                Console.WriteLine( "Error: " + ex.Message );
            }
            finally
            {
                RRUtil.DisposeOfAllRedRats();
            }
        }

        /// <summary>
        /// Simple check of program arguments. We need all three in the correct order to work.
        /// </summary>
        private void CheckArgs( string[] args )
        {
            // Need all three arguments.
            if ( ( args != null ) && ( args.Length == 3 ) ) return;
            throw new Exception( "Invalid argument list.\n\nUsage: SimpleSignalOutput database-file.xml device-name signal-name\n");
        }

        /// <summary>
        /// Simply finds the first USB RedRat attached to this computer.
        /// </summary>
        private IRedRat3 FindRedRat()
        {
            if (!(RRUtil.GetDefaultUsbRedRat() is IRedRat3 rr3))
            {
                throw new Exception("Unable to find any USB RedRat devices attached to this computer.");
            }

            return rr3;
        }

        /// <summary>
        /// Returns an IR signal object from the signal DB file using the deviceName and signalName to
        /// look it up.
        /// </summary>
        private IRPacket GetSignal( AVDeviceDB signalDB, string deviceName, string signalName )
        {
            var device = signalDB.GetAVDevice( deviceName );
            if ( device == null )
            {
                throw new Exception( $"No device of name '{deviceName}' found in the signal database." );
            }
            var signal = device.GetSignal( signalName );
            if ( signal == null )
            {
                throw new Exception( $"No signal of name '{signalName}' found for device '{deviceName}' in the signal database." );
            }
            return signal;
        }
    }
}
