﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using RedRat.AvDeviceDb;
using RedRat.IR;
using RedRat.RedRat4;
using RedRat.RedRat4.StreamIO.IP;
using RedRat.Util;
using RedRat.Util.Serialization;

namespace RedRat.Samples.IrNetBoxDemo
{
    /// <summary>
    /// Example code that shows how to use the RedRat API to communicate with an irNetBox.
    /// 
    /// When compiled, run as: "irNetBoxDemo.exe [IP address | MAC address | name]". For example:
    /// 
    ///    > irNetBoxDemo.exe 192.168.1.40
    ///    > irNetBoxDemo.exe 00-20-4A-CF-2B-82
    ///    > irNetBoxDemo.exe "irNB Rack-1"
    /// 
    /// This code shows three ways of IR output via the irNetBox:
    /// 
    /// 1) Using the old API in an identical fashion to the MK-II box. One IR signal is sent
    ///    at a time and there is no speed improvement.
    /// 
    /// 2) Using the Async Output Manager. This is basically a queue that adds as many IR signals
    ///    as needed and sends them on to the irNetBox-III/IV as quickly as possible. It is simple
    ///    to use and does the clever stuff of ensuring that IR signals are not sent to busy output
    ///    ports. The disadvantage is that signals are queued, so the exact timing of output is not
    ///    controllable. In most practical situations, the IR signal transmission is instantaneous,
    ///    so simplicity of use probably trumps exactness.
    /// 
    /// 3) Directly using the new async API. This is more complex but gives better control of the
    ///    box. The example here starts one worker task for each output port and sends IR signals
    ///    to that port as fast as possible.
    ///
    /// Runs on Windows and Linux (.NET Core).
    /// 
    /// </summary>
    public class IRNetBoxDemo
    {
        /// <summary>
        /// Use this IP address if none given as first command line args.
        /// </summary>
        public const string defaultIPAddress = "192.168.86.30";

        private readonly Random rand = new Random();
        public const string dbFilename = "DeviceDB.xml";
        private AVDeviceDB deviceDB;
        private readonly object fileLoadLock = new object();

        private static void Main( string[] args )
        {
            try
            {
                new IRNetBoxDemo( args );
            }
            catch ( Exception ex )
            {
                Console.Error.WriteLine( ex );
            }
            finally
            {
                Console.WriteLine( "Hit <RETURN> to exit..." );
                Console.ReadLine();
            }
        }

        /// <summary>
        /// .ctor
        /// </summary>
        public IRNetBoxDemo( string[] args )
        {
            Console.WriteLine( "\nIRNetBoxDemo\n=================\n" );

            // The first job is to find the irNetBox we want.
            var locationInfo = FindIrNetBox( args.Length > 0 ? args[0].Trim() : defaultIPAddress );
            if ( locationInfo == null )
            {
                return;
            }

            // Get a generic irNetBox (IRedRat4) object to use
            using ( var irNB = locationInfo.GetRedRat() as IRedRat4 )
            {

                // Check that the cast has worked ok. It should have worked fine this case as the original search 
                // was just for these type of RedRat devices, but good practice in case this code is used elsewhere.
                if ( irNB == null )
                {
                    Console.WriteLine( "The irNetBox with IP/MAC/name of '{0}' is not of type MK-III or MK-IV.", args[0] );
                    return;
                }

                // Connect to irNetBox.
                irNB.Connect();

                // Blink a few times...
                Console.WriteLine( "Blinking..." );
                for ( var i = 0; i < 4; i++ )
                {
                    irNB.Blink();
                    Thread.Sleep( 500 );
                }

                // ===== Do work ====
                // 1. Using synchronous API. No faster than the MK-II irNetBox.
                DoSyncApiCalls( irNB );

                // Check to see if this is a MK-III (or greater) irNetBox
                if ( LocationInfo.IsMKIII( irNB.LocationInformation.RRType ) || LocationInfo.IsMKIV( irNB.LocationInformation.RRType ) )
                {
                    // Cast as a MK-III as this gives us the async API to use.
                    var irNBIII = irNB as IRedRat4_III;

                    // 2. Using the async output manager. Easy but less control.
                    DoAsyncOutputManagerCalls( irNBIII );

                    // 3. Using the async API. More involved (not much) but more control.
                    DoAsyncApiCalls( irNBIII );
                }

                // Disconnect from box
                irNB.Disconnect();
            }
        }

        /// <summary>
        /// Given an IP address, MAC address or friendly name, find the irNetBox we want.
        /// </summary>
        public LocationInfo FindIrNetBox( string irNetBoxId )
        {
            // If the IP address is dynamically assigned, then we may not know what it is, so we should use:
            //    1) The MAC address as given on the back of the irNetBox.
            //    2) A friendly name, which can be assigned with a RedRat application such as the 
            //       Signal DB Utility (RedRats menu item).

            // Do a network search (broadcast) to find any irNetBoxes out there. This will take a few seconds to return.
            // The search returns a list of LocationInfo objects which are used to provide information about the irNetBoxes.
            // We can cast them all to type RedRat4IpLocationInfo as we are only searching for IP irNetBoxes.
            Console.WriteLine( "Searching for irNetBoxes...." );
            var irNBLocInfos = RRUtil.FindRedRats( LocationInfo.RedRatType.RedRat4 ).Cast<RedRat4IpLocationInfo>();

            // Print the list...
            if ( !irNBLocInfos.Any() )
            {
                Console.WriteLine( "No irNetBoxes found on network." );
                return null;
            }

            Console.WriteLine( "Found the following:" );
            foreach ( var locationInfo in irNBLocInfos )
            {
                Console.WriteLine( $"\t{locationInfo.IPAddress}" +
                                   $"\t{IpUtils.MacAddrBytesToStr( locationInfo.MACAddress )}" +
                                   $"\t{locationInfo.RRType}\t'{locationInfo.Name}'" );
            }

            Console.WriteLine();

            // Check IP address...
            var irNbLocationInfo = irNBLocInfos.FirstOrDefault( rr => rr.IPAddress.ToString().Equals( irNetBoxId ) );
            if ( irNbLocationInfo != null )
            {
                Console.WriteLine( $"Found irNetBox with IP = {irNbLocationInfo.IPAddress}" );
                return irNbLocationInfo;
            }

            // Check MAC address...
            irNbLocationInfo = irNBLocInfos.FirstOrDefault( rr => IpUtils.MacAddrBytesToStr( rr.MACAddress ).Equals( irNetBoxId ) );
            if ( irNbLocationInfo != null )
            {
                Console.WriteLine( $"Found irNetBox with MAC address = '{IpUtils.MacAddrBytesToStr( irNbLocationInfo.MACAddress )}'" );
                return irNbLocationInfo;
            }

            // Check friendly name...
            irNbLocationInfo = irNBLocInfos.FirstOrDefault( rr => rr.Name.Equals( irNetBoxId ) );
            if ( irNbLocationInfo != null )
            {
                Console.WriteLine( $"Found irNetBox with name = '{irNbLocationInfo.Name}'" );
                return irNbLocationInfo;
            }

            // Can't find the irNetBox.
            Console.WriteLine( $"No irNetBox found with IP/MAC/Name of {irNetBoxId}" );
            return null;
        }

        /// <summary>
        /// Use the standard synchronous API. This works for all types of irNetBox, but it is quite slow as
        /// it only sends one IR signal at a time. For MK-III (or higher) irNetBoxes, the async API can be used instead.
        /// </summary>
        public void DoSyncApiCalls( IRedRat4 irNB )
        {
            Console.WriteLine( "\nSYNC API CALLS" );

            // Output an IR signal via a few enabled outputs
            Console.WriteLine( "\tOutputting a couple of IR signals..." );
            var opState = new OutputList
            {
                new OutputPort( 2, OutputListUtil.OutputPowerToInt( OutputPower.LOW ) ),
                new OutputPort( 4, OutputListUtil.OutputPowerToInt( OutputPower.LOW ) ),
                new OutputPort( 5, OutputListUtil.OutputPowerToInt( OutputPower.LOW ) )
            };
            irNB.EnabledIROutputs = opState; // Set the output state.
            for ( var i = 0; i < 4; i++ )
            {
                // Send the IR signal to the enabled outputs.
                irNB.OutputModulatedSignal( GetSignal() );
                Thread.Sleep( 500 );
            }

            Console.WriteLine( "\tDone." );
        }

        /// <summary>
        /// Output IR signals using the async output manager. This is basically a queue into which IR
        /// signals are inserted and they are output as quickly as possible. The disadvantage is that
        /// one doesn't get notified when a particular IR signal output is complete - for that one
        /// needs the normal async API.
        /// </summary>
        public void DoAsyncOutputManagerCalls( IRedRat4_III irNBIII )
        {
            Console.WriteLine( "\nASYNC OUTPUT MANAGER CALLS" );

            // Display info if MK-IV
            var irNBIV = irNBIII as IRedRat4_IV;
            irNBIV?.DisplayText( "Async o/p V1" );

            // Get the Async Output Manager.
            var asyncOutputManager = irNBIII.AsyncOutputManager;

            const int numberOfSignals = 200; // Output this many IR signals.

            // Send a load of IR signals to a random selection of ports.
            Console.WriteLine( $"\tSending {numberOfSignals} IR signals via the async output manager..." );
            for ( var i = 0; i < numberOfSignals; i++ )
            {
                // Queue the signals for output. Just use one output port for each
                // signal, though we can send to more if wanted.
                var opList = new OutputList
                {
                    new OutputPort
                    {
                        PortNumber = rand.Next( 1, irNBIII.NumIROutputs + 1 ),
                        PowerLevel = 50, // Half power
                    }
                };

                // Queue each signal for output.
                asyncOutputManager.QueueSignal( GetSignal(), opList );
            }

            // Wait for completion.
            // For simplicity we just poll the size of the queue to see when it has finished. In a real world application
            // this would not be necessary. A 'better' way of doing this would be to use callbacks from the QueueSignal
            // method call to monitor when IR signal output is complete.
            while ( asyncOutputManager.QueueState.Length > 0 )
            {
                Thread.Sleep( 250 );
            }

            irNBIV?.DisplayText( "" );

            Console.WriteLine( "\tDone." );

            // Wait for the last IR signal transmission to complete.
            Thread.Sleep( 500 );
        }

        /// <summary>
        /// Uses the standard async API to send IR signals to different outputs.
        /// </summary>
        public void DoAsyncApiCalls( IRedRat4_III irNBIII )
        {
            const int numberOfSignals = 40; // Output this many IR signals on each port.

            // Display info if MK-IV
            var irNBIV = irNBIII as IRedRat4_IV;
            if ( irNBIV == null )
            {
                Console.Error.WriteLine( "ERROR: Device is not a MK-IV irNetBox." );
                return;
            }

            irNBIV.DisplayText( "Async o/p V2" );

            Console.WriteLine( "\nASYNC API CALLS" );
            Console.WriteLine( $"\tStarting {irNBIII.NumIROutputs} tasks to output {numberOfSignals} IR signals on each port." );

            const int outputPowerLevel = 45;

            var tasks = new List<Task>();

            // What we do here is start a new thread for each port, in effect having 16 different
            // entities sending IR data at the same time.
            for ( var op = 0; op < irNBIII.NumIROutputs; op++ )
            {
                var thisOutput = op + 1;    // Prevent access to modified closure.
                var t = Task.Run( () => SignalOutputTaskAsync( irNBIII, thisOutput, outputPowerLevel, numberOfSignals ) );
                tasks.Add( t );
            }

            // Wait for all tasks to complete.
            Task.WaitAll( tasks.ToArray() );

            irNBIV.DisplayText( "" );

            Console.WriteLine( "\tDone." );
        }

        /// <summary>
        /// Task loop sending IR signals via a single port.
        /// </summary>
        public async Task SignalOutputTaskAsync(IRedRat4_III rr4, int port, int powerLevel, int numberToOutput)
        {
            try
            {
                // Setup output list information for the IR output call. In this case it is a single
                // output - the port that this thread is responsible for.
                var opList = new OutputList {new OutputPort( port, powerLevel )};
                var opLock = new SemaphoreSlim( 0, 1 );

                // Pause after each IR signal. Needed for the receiver to treat as separate.
                const int postSigDelay = 100;

                for ( var i = 0; i < numberToOutput; i++ )
                {
                    // Send the IR signal. If this port is busy then an error is returned, so we have to wait until
                    // the previous IR signal output has completed. We are notified of completion in the callback.
                    await Task.Run( async () =>
                    {
                        rr4.OutputIRPacketAsync( GetSignal(), opList, postSigDelay, null, asyncResult =>
                        {
                            // This is the callback once the IR signal has been transmitted by the irNetBox, plus any post
                            // signal delay.

                            // Do we have an error returned?
                            if ( asyncResult.Error != null )
                            {
                                Console.WriteLine( asyncResult.Error );
                            }

                            // Signal that the output task can continue with the next IR signal.
                            opLock.Release();
                        });

                        // Wait until the IR signal transmission has been completed.
                        await opLock.WaitAsync();
                    } );

                    // Don't really need this, but reduce speed a bit to help see what's going on the box's display.
                    await Task.Delay( 250 );
                }
            }
            catch ( Exception ex )
            {
                Console.WriteLine( ex );
            }
            finally
            {
                Console.WriteLine( $"\t>> Worker task for port {port} complete." );
            }
        }

        /// <summary>
        /// Returns a random IR signal for output.
        /// </summary>
        public ModulatedSignal GetSignal()
        {
            // Load device DB containing IR signals if have not already done so.
            // Lock to ensure thread safety. If the async tasks are the first ones run then it multiple
            // tasks will demand access to the DB and so try and load it.
            lock ( fileLoadLock )
            {
                if ( deviceDB == null )
                {
                    deviceDB = Serializer.AvDeviceDbFromXmlFile(dbFilename).Object;
                }
            }

            // Get a random IR signal
            ModulatedSignal signal = null;
            // Not all IR signals read from the DB are modulated signals, so loop until we get one.
            while ( signal == null )
            {
                var index = rand.Next( deviceDB.AVDevices.Length );
                var device = deviceDB.AVDevices[index];
                signal = device.Signals[rand.Next( device.Signals.Length )] as ModulatedSignal;
            }

            return signal;
        }
    }
}
